from unittest import TestCase
import gevent
from gevent.queue import Queue
import datetime
import mock

from logcabin.event import Event
from logcabin.context import DummyContext
from logcabin.filters import json, mutate, python, regex, stats, syslog, url

from testhelper import assertEventEquals, about, between

class FilterTests(TestCase):
    def create_stage(self, **conf):
        return self.cls(**conf)

    def create(self, conf={}, events=[]):
        self.input = Queue()
        self.output = Queue()
        with DummyContext():
            self.i = self.create_stage(**conf)
        self.input = self.i.setup(self.output)

        self.i.start()
        for ev in events:
            self.input.put(ev)
        return self.i

    def wait(self, timeout=1.0, events=1):
        with gevent.Timeout(timeout):
            # wait for input to be consumed and output to be produced
            while self.input.qsize():
                gevent.sleep(0.0)
            while self.output.qsize() < events:
                gevent.sleep(0.0)

        if events:
            return [self.output.get() for n in xrange(events)]

    def tearDown(self):
        self.i.stop()

class JsonTests(FilterTests):
    cls = json.Json

    def test_consume(self):
        self.create({},
            [Event(data='{"a": 1}')])
        q = self.wait()
        assertEventEquals(self, Event(a=1), q[0])

    def test_consume_false(self):
        self.create({'consume': False},
            [Event(data='{"a": 1}')])
        q = self.wait()
        assertEventEquals(self, Event(a=1, data='{"a": 1}'), q[0])

    def test_bad_json(self):
        self.create({'consume': False},
            [Event(data='"invalid')])
        self.wait(events=0)
        self.assertEquals(0, self.output.qsize())

class RegexTests(FilterTests):
    cls = regex.Regex

    def test_match(self):
        self.create({'regex': r'(?P<letters>[a-z]+)(?P<numbers>\d+)'},
            [Event(data='abc123')])
        q = self.wait()
        assertEventEquals(self, Event(letters='abc', numbers='123'), q[0])

    def test_no_match(self):
        self.create({'regex': r'(?P<letters>[a-z]+)(?P<numbers>\d+)',
            'on_error': 'tag'},
            [Event(data='.!$#')])
        q = self.wait()
        self.assertEquals(['error'], q[0].tags)

class MutateTests(FilterTests):
    cls = mutate.Mutate

    def test_set(self):
        self.create({'set': {'a': 2}},
            [Event(a=1)])
        q = self.wait()
        assertEventEquals(self, Event(a=2), q[0])

    def test_set_format(self):
        self.create({'set': {'a': '{b}.{c}'}},
            [Event(b=1, c=2)])
        q = self.wait()
        assertEventEquals(self, Event(a='1.2', b=1, c=2), q[0])

    def test_rename(self):
        self.create({'rename': {'b': 'a', 'd': 'c'}},
            [Event(a=1)])
        q = self.wait()
        assertEventEquals(self, Event(b=1), q[0])

    def test_copy(self):
        self.create({'copy': {'b': 'a', 'd': 'c'}},
            [Event(a=1, c=5)])
        q = self.wait()
        assertEventEquals(self, Event(b=1, a=1, c=5, d=5), q[0])

    def test_unset(self):
        self.create({'unset': ['a', 'c']},
            [Event(a=1, b=2)])
        q = self.wait()
        assertEventEquals(self, Event(b=2), q[0])

class PythonTests(FilterTests):
    cls = python.Python

    def test_simple(self):
        function = mock.Mock()
        ev = Event(data='abc123')
        self.create({'function': function},
            [ev])
        q = self.wait()
        
        assertEventEquals(self, Event(data='abc123'), q[0])
        function.assert_called_with(ev)

class UrlTests(FilterTests):
    cls = url.Url

    def test_match(self):
        self.create({'field': 'data'},
            [Event(data='abc?a=1%2C2&b=2+3')])
        q = self.wait()
        assertEventEquals(self, Event(path='abc', a='1,2', b='2 3'), q[0])

    def test_no_match(self):
        self.create({'field': 'data',
            'on_error': 'tag'},
            [Event(data=None)])
        q = self.wait()
        self.assertEquals(['error'], q[0].tags)

class StatsTests(FilterTests):
    cls = stats.Stats

    events = [
        Event(controller='home', action='index', duration=3.0, bytes=6926, timings={'view': 1.0}),
        Event(controller='home', action='login', duration=2.4, bytes=15568, timings={'view': 2.0}),
        Event(controller='home', action='index', duration=4.0, bytes=18150, timings={'view': 1.2}),
        Event(controller='home', action='index', duration=3.5, bytes=30159, timings={'view': 2.3}),
        Event(controller='missing', action='duration'),
        Event(someotherevent='blah', duration=3.5),
    ]

    def test_nested(self):
        self.create({'period': 0.1, 'metrics': {'rails.{controller}.{action}.{0}': 'timings.*'}},
            self.events)

        # 8 events expected - the above 6, and then 2 stat events
        q = self.wait(events=8)

        q = [i for i in q if i.stats]
        q.sort(key=lambda k: k.metric)

        expected = Event(metric='rails.home.index.timings.view',
            stats={
                'count': 3,
                'max': 2.3,
                'mean': 1.5,
                'median': 1.2,
                'min': 1.0,
                'rate': between(1, 100),
                'stddev': about(2.34, 2),
                'upper95': 2.19,
                'upper99': 2.278},
            tags=['stat'],
        )
        assertEventEquals(self, expected, q[0])

        expected = Event(metric='rails.home.login.timings.view',
            stats={
                'count': 1,
                'max': 2.0,
                'mean': 2.0,
                'median': 2.0,
                'min': 2.0,
                'rate': between(1, 100),
                'stddev': 0.0,
                'upper95': 2.0,
                'upper99': 2.0},
            tags=['stat'],
        )
        assertEventEquals(self, expected, q[1])

        # wait for further two 'zero' stats
        q = self.wait(events=2)

    def test_wildcard(self):
        self.create({'period': 0.1, 'metrics': {'rails.{controller}.{action}.{0}': '*'}},
            self.events)

        # 8 events expected - the above 6, and then 4 stat events
        q = self.wait(events=10)

        q = [i for i in q if i.stats]
        q.sort(key=lambda k: k.metric)

        expected = Event(metric='rails.home.index.bytes',
            stats={
                'count': 3,
                'rate': between(1, 100),
                'max': 30159,
                'min': 6926,
                'median': 18150,
                'mean': 18411,
                'stddev': about(30789),
                'upper95': 28958.1,
                'upper99': 29918.82,
            },
            tags=['stat'],
        )
        assertEventEquals(self, expected, q[0])

        expected = Event(metric='rails.home.index.duration',
            stats={
                'count': 3,
                'rate': between(1, 100),
                'max': 4.0,
                'min': 3.0,
                'median': 3.5,
                'mean': 3.5,
                'stddev': 5.0,
                'upper95': 3.95,
                'upper99': 3.99,
            },
            tags=['stat'],
        )
        assertEventEquals(self, expected, q[1])

        expected = Event(metric='rails.home.login.duration',
            stats={
                'count': 1,
                'rate': between(1, 100),
                'max': 2.4,
                'min': 2.4,
                'median': 2.4,
                'mean': 2.4,
                'stddev': 0.0,
                'upper95': 2.4,
                'upper99': 2.4,
            },
            tags=['stat'],
        )
        assertEventEquals(self, expected, q[3])

class SyslogTests(FilterTests):
    cls = syslog.Syslog

    # Formats - you can never have enough of them!
    good_packets = [
        '<174>Nov 30 19:56:13 host01 prog[1234]: log message', # RSYSLOG_ForwardFormat
        '<174>Mar  4 11:57:46 micro01 testlog.py: test', # RSYSLOG_TraditionalFileFormat
        '<174>2012-12-07T13:44:27.710956+01:00 test01 program: test' # RSYSLOG_ForwardFormat
    ]
    good_events = [
        Event(
            timestamp=datetime.datetime(2013, 10, 30, 19, 56, 13),
            facility='local5',
            severity='Informational',
            host='host01',
            program='prog',
            pid='1234',
            message='log message'),
        Event(
            timestamp=datetime.datetime(2013, 3, 4, 11, 56, 46),
            facility='local5',
            severity='Informational',
            host='micro01',
            program='testlog.py',
            pid=None,
            message='test'),
        Event(
            timestamp=datetime.datetime(2012, 12, 7, 12, 44, 27, 710956),
            facility='local5',
            severity='Informational',
            host='test01',
            program='program',
            pid=None,
            message='test')

    ]
    bad_packets = ['<>Nov 30 19:56:13 host01 prog[1234]: log message']

    def test_good(self):
        self.create({},
            [Event(data=x) for x in self.good_packets])

        events = self.wait(events=len(self.good_events))
        for ev in self.good_events:
            assertEventEquals(self, ev, events.pop(0))

    def test_bad(self):
        self.create({'consume': False, 'on_error': 'tag'},
            [Event(data=x) for x in self.bad_packets])
        events = self.wait(events=len(self.bad_packets))

        bad_events = [Event(data=x, message='invalid syslog', tags=['error']) for x in self.bad_packets]
        for ev in bad_events:
            assertEventEquals(self, ev, events.pop(0))



from .filter import Filter

class Python(Filter):
    """Call out to a python function for adding custom functionality.

    :param callable function: callable taking the event as an argument
    """

    def __init__(self, function, on_error='reject'):
        super(Python, self).__init__(on_error=on_error)
        self.function = function

    def process(self, event):
        return self.function(event)



import math

def mean(li):
    """
    Calculate mean of a list.

    >>> mean([0.5,2.0,3.5])
    2.0
    >>> mean([])
    """
    if li:
        return sum(li) / len(li)
    return None

def percentile(li, pc):
    """
    Calculate percentiles of a list.

    >>> li = [0,1,2,3,4,5,6,7,8,9,10]
    >>> percentile(li, 0.5)
    5
    >>> percentile(li, 0.9)
    9
    >>> percentile(li, 1.0)
    10
    >>> percentile(li, 0.0)
    0
    >>> percentile([], 0.5)

    :param list li: data points
    :param float pc: percentile from 0.0 to 1.0
    """
    if not li:
        return None
    k = (len(li)-1) * pc
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return li[int(k)]
    d0 = li[int(f)] * (c-k)
    d1 = li[int(c)] * (k-f)
    return d0 + d1

def stddev(li, mean):
    """
    Calculate the standard deviation of a set of data.

    >>> stddev([3, 3.5, 4], 3.5)
    5.0
    """
    if not li:
        return None
    return math.sqrt(sum(x*x for x in li) - mean*mean)



from unittest import TestCase
import gevent
from gevent.queue import Queue
import gevent.server
import mock
import json
import gzip
import os
import random
import struct
import time
from datetime import datetime
import cPickle as pickle
import zmq.green as zmq

from logcabin.event import Event
from logcabin.context import DummyContext

from logcabin.outputs import elasticsearch, file as fileoutput, graphite, log, \
    mongodb, perf, s3, zeromq

from testhelper import TempDirectory, assertEventEquals

class OutputTests(TestCase):
    def create(self, conf={}):
        with DummyContext():
            self.i = i = self.cls(**conf)
        self.output = Queue()
        self.input = i.setup(self.output)
        i.start()
        return i

    def waitForEmpty(self, timeout=1.0):
        with gevent.Timeout(timeout):
            while self.input.qsize():
                gevent.sleep()

    def tearDown(self):
        self.i.stop()

class LogTests(OutputTests):
    cls = log.Log

    def test_log(self):
        self.create()

        self.input.put(Event(field='x'))
        self.waitForEmpty()

        self.assertEquals(0, self.input.qsize())

class ElasticsearchTests(OutputTests):
    cls = elasticsearch.Elasticsearch

    def test_log(self):
        with mock.patch('urllib2.urlopen') as urlopen_mock:
            urlopen_mock.return_value.read.return_value = json.dumps({'_type': 'event', '_id': 'w0HnGYHFSOS7EBIWnxBcEg', 'ok': True, '_version': 1, '_index': 'test'})
            i = self.create({'index': 'test', 'type': 'event'})

            self.input.put(Event(field='x'))
            self.waitForEmpty()
            i.stop()

            self.assert_(urlopen_mock.called)

class FileTests(OutputTests):
    cls = fileoutput.File

    def assertFileContents(self, expected, filename):
        if filename.endswith('.gz'):
            actual = gzip.open(filename).read()
        else:
            actual = file(filename).read()

        self.assertEquals(expected, actual)

    events = [Event(program='httpd'), Event(program='ntpd')]

    def test_simple(self):
        with TempDirectory():
            self.create({'filename': 'log/output_{program}.log'})

            map(self.input.put, self.events)
            self.waitForEmpty()

            self.assertFileContents(self.events[0].to_json()+'\n', 'log/output_httpd.log')
            self.assertFileContents(self.events[1].to_json()+'\n', 'log/output_ntpd.log')

    def test_max_size(self):
        with TempDirectory():
            self.create({'filename': 'output.log',
                'max_size': 16})

            map(self.input.put, self.events)
            self.waitForEmpty()

            self.assertFileContents(self.events[0].to_json()+'\n', 'output.log.1')
            self.assertFileContents(self.events[1].to_json()+'\n', 'output.log')

            # assert the 'fileroll' event is generated
            self.assert_(self.output.qsize())
            events = [self.output.get() for i in xrange(self.output.qsize())]
            assertEventEquals(self, Event(tags=['fileroll'], filename='output.log.1', last=self.events[0], trigger=self.events[1]), events[1])

    def test_timestamped(self):
        events = [
            Event(message='1', timestamp=datetime(2013, 1, 1, 23, 59, 58)),
            Event(message='2', timestamp=datetime(2013, 1, 1, 23, 59, 59)),
            Event(message='3', timestamp=datetime(2013, 1, 2, 0, 0, 0)),
            Event(message='4', timestamp=datetime(2013, 1, 2, 0, 0, 1))]

        with TempDirectory():
            self.create({'filename': 'output-{timestamp:%Y%m%d}.log',
                'compress': 'gz'})

            map(self.input.put, events)
            self.waitForEmpty()

            self.assertFileContents(events[0].to_json()+'\n'+events[1].to_json()+'\n', 'output-20130101.log.1.gz')
            self.assertFileContents(events[2].to_json()+'\n'+events[3].to_json()+'\n', 'output-20130102.log')

            # assert the 'fileroll' event is generated
            self.assertEquals(len(events)+1, self.output.qsize())
            outputs = [self.output.get() for i in xrange(self.output.qsize())]
            assertEventEquals(self, Event(tags=['fileroll'], filename='output-20130101.log.1.gz', last=events[1], trigger=events[2]), outputs[2])

    def test_compress(self):
        with TempDirectory():
            self.create({'filename': 'output.log',
                'max_size': 16,
                'compress': 'gz'})

            map(self.input.put, self.events)
            self.waitForEmpty()
            self.i.stop()

            self.assertFileContents(self.events[0].to_json()+'\n', 'output.log.1.gz')
            self.assertFileContents(self.events[1].to_json()+'\n', 'output.log')

    def test_max_count(self):
        with TempDirectory():
            self.create({'filename': 'output.log',
                'max_size': 16,
                'max_count': 2})

            map(self.input.put, self.events * 10)
            self.waitForEmpty()

            self.assertFileContents(self.events[1].to_json()+'\n', 'output.log')
            self.assert_(not os.path.exists('output.log.3'))

class PerfTests(OutputTests):
    cls = perf.Perf

    def test_log(self):
        self.create()

        self.input.put(Event(field='x'))
        self.waitForEmpty()

        self.assertEquals(0, self.input.qsize())

class MongodbTests(OutputTests):
    cls = mongodb.Mongodb

    @mock.patch('pymongo.MongoClient')
    def test_log(self, mock_client):
        mock_db = mock_client.return_value.__getitem__ = mock.Mock()
        mock_col = mock_db.return_value.__getitem__ = mock.Mock()
        mock_col.return_value.insert.return_value = 'abc123' # mock id return

        self.create()

        self.input.put(Event(field='x'))
        self.waitForEmpty()

        self.assertEquals(0, self.input.qsize())

class S3Tests(OutputTests):
    cls = s3.S3

    @mock.patch('boto.connect_s3')
    def test_upload(self, mock_s3):
        mock_conn = mock_s3.return_value
        mock_bucket = mock_conn.get_bucket.return_value
        mock_key = mock_bucket.new_key.return_value

        with TempDirectory():
            # create log file for it to upload
            with file('output.log', 'w') as fout:
                fout.write('a log file')

            self.create({'access_key': 'dummy', 'secret_key': 'dummy',
                'bucket': 'bucket1', 'path': 'logs/1.json'})
            self.input.put(Event(tags=['fileroll'], filename='output.log'))
            self.waitForEmpty()

        mock_s3.assert_called_with('dummy', 'dummy')
        mock_conn.get_bucket.assert_called_with('bucket1')
        mock_bucket.new_key.assert_called_with('logs/1.json')
        mock_key.set_contents_from_filename.assert_called_with('output.log')

class GraphiteTests(OutputTests):
    cls = graphite.Graphite

    @mock.patch('logcabin.event.datetime')
    def test_log(self, mock_datetime):
        now = datetime(2013, 1, 1, 2, 34, 56, 789012)
        t_now = int(time.mktime(now.timetuple()))
        mock_datetime.utcnow.side_effect = lambda: now

        port = random.randint(1024, 65535)
        received = []
        
        def handle(socket, address):
            # graphite is a 4-byte length, followed by pickled representation
            length, = struct.unpack('!L', socket.recv(4))
            d = socket.recv(length)
            received.append(pickle.loads(d))
        server = gevent.server.StreamServer(('', port), handle)
        server.start()

        self.create({'port': port})

        self.input.put(Event(metric='a.b.c', stats={'mean': 1.5, 'min': 1.0}))
        self.waitForEmpty()
        self.i.stop()

        self.assertEquals(0, self.input.qsize())
        server.stop()
        self.assertEquals(1, len(received))
        self.assertEquals([
            ('a.b.c.min', (t_now, 1.0)),
            ('a.b.c.mean', (t_now, 1.5))
        ], received[0])

    def test_unavailable(self):
        port = random.randint(1024, 65535)
        received = []
        
        def handle(socket, address):
            # graphite is a 4-byte length, followed by pickled representation
            length, = struct.unpack('!L', socket.recv(4))
            d = socket.recv(length)
            received.append(pickle.loads(d))
        server = gevent.server.StreamServer(('', port), handle)

        self.create({'port': port})

        self.input.put(Event(metric='a.b.c', stats={'mean': 1.5, 'min': 1.0}))
        server.start()

        self.waitForEmpty()

        self.assertEquals(0, self.input.qsize())
        self.i.stop()
        server.stop()
        self.assertEquals(1, len(received))

class ZeromqTests(OutputTests):
    cls = zeromq.Zeromq

    def test_log(self):
        conf = {'address': 'ipc://testipc2', 'mode': 'connect', 'socket': 'PUSH'}
        self.create(conf)

        # create a zeromq socket
        ctx = zmq.Context()
        sock = ctx.socket(zmq.PULL)
        sock.bind(conf['address'])

        ev = Event(field='x')
        self.input.put(ev)
        self.waitForEmpty()

        data = sock.recv()
        self.assertEquals(ev.to_json(), data)



from .output import Output
import boto
import boto.s3.key

class S3(Output):
    """Uploads to an S3 bucket.

    This should follow the File output, in order to upload the rolled log files
    to S3::

        File(filename='log/batch.log', )
        If(lambda ev: 'fileroll' in ev.tags):
            S3(access_key='...', secret_key='...',
               bucket='x', path='logs/{timestamp:%Y-%m-%d/%H%M%S}.log')

    Bucket or path may by formatted by event, eg. to upload to a timestamped path:
    path='{timestamp:%Y-%m-%d/%H%M%S}.log'

    :param string access_key: Amazon S3 access key
    :param string secret_key: Amazon S3 secret key
    :param string bucket: the bucket name
    :param string path: the path
    """

    def __init__(self, access_key, secret_key, bucket, path):
        super(S3, self).__init__()
        self.access_key = access_key
        self.secret_key = secret_key
        self.bucket = bucket
        self.path = path

    def process(self, event):
        bucket = event.format(self.bucket)
        path = event.format(self.path)

        filename = event['filename']
        self.logger.info('Uploading %s to s3://%s/%s' % (filename, bucket, path))
        s3 = boto.connect_s3(self.access_key, self.secret_key)
        bucket = s3.get_bucket(bucket)
        k = bucket.new_key(path)
        k.set_contents_from_filename(filename)



from .output import Output
import zmq.green as zmq

class Zeromq(Output):
    """Outputs on a zeromq socket.

    :param string address: zeromq address (default: `tcp://*:2120`)
    :param string mode: connect or bind (default: connect)
    :param string socket: PUSH or PUB (default: PUSH)
    """
    def __init__(self, address='tcp://*:2120', mode='connect', socket='PUSH'):
        if mode not in ('connect', 'bind'):
            raise ValueError('mode should be connect or bind')
        if socket not in ('PUSH', 'PUB'):
            raise ValueError('socket should be PUSH or PUB')

        super(Zeromq, self).__init__()
        self.ctx = zmq.Context()
        self.sock = self.ctx.socket(getattr(zmq, socket))
        self.address = address
        if mode == 'connect':
            self.sock.connect(self.address)
        else:
            self.sock.bind(self.address)

    def process(self, event):
        data = event.to_json()
        self.sock.send(data)



from unittest import TestCase

from logcabin.configuration import PyConfigLoader, ConfigException

import os

def _p(p):
    return os.path.join(os.path.dirname(__file__), p)

class PyConfigLoaderTests(TestCase):
    def test_good(self):
        loader = PyConfigLoader(_p('config/good.py'))
        loader.configure()
        pipeline = loader.pipeline

        # kick the tyres
        pipeline.start()
        pipeline.stop()

    def test_complex(self):
        loader = PyConfigLoader(_p('config/complex.py'))
        loader.configure()
        pipeline = loader.pipeline

        # kick the tyres
        pipeline.start()
        import gevent
        gevent.sleep()
        pipeline.stop()

    def test_bad(self):
        loader = PyConfigLoader(_p('test/config/bad.py'))
        self.assertRaises(ConfigException, loader.configure)



#!/usr/bin/env python

from setuptools import setup, find_packages
from setuptools import Command

class tag(Command):
    """Tag git release."""

    description = __doc__
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import logcabin
        import versiontools
        import subprocess
        version = versiontools.format_version(logcabin.__version__)
        ret = subprocess.call(['git', 'tag', '-a', version, '-m', version])
        if ret:
            raise SystemExit("git tag failed")
        ret = subprocess.call(['git', 'push', '--tags'])
        if ret:
            raise SystemExit("git push --tags failed")

setup(
    name="logcabin",
    url='http://github.com/artirix/logcabin/',
    version=':versiontools:logcabin:',
    license='Apache 2.0',
    description="Logging framework for receiving and processing events from a "
    "diverse range of sources and formats, and relaying onto multiple "
    "destinations.",
    long_description=file('README.rst').read(),
    author='Barnaby Gray',
    author_email='barnaby@artirix.com',
    setup_requires=[
        'versiontools',
    ],
    install_requires=[
        'python-dateutil',
        'gevent >= 0.13.6',
    ],
    tests_require=[
        'nose',
    ],
    test_suite="nose.collector",
    packages=find_packages(exclude=('docs', 'test')),
    package_data={'logcabin': ['logging.cfg']},
    zip_safe=False,
    entry_points={
        'console_scripts':
        ['logcabin = logcabin:main']
    },
    cmdclass={
        'tag': tag,
    },
)



import time
import gevent

class ConfigException(Exception):
    pass

def dynamic_class(modname, classname):
    mod = __import__(modname)
    # import returns the top-level module, so recurse to the sub module
    for m in modname.split('.')[1:]:
        mod = getattr(mod, m)
    return getattr(mod, classname)

class BroadcastQueue(list):
    """Queue-like object that broadcasts to all child queues."""

    def put(self, obj):
        for q in self:
            q.put(obj)

    def join(self):
        for q in self:
            q.join()

class Periodic(gevent.Greenlet):
    """Greenlet wrapper that periodically makes a callback."""
    
    def __init__(self, period, callback):
        super(Periodic, self).__init__()
        self.period = period
        self.callback = callback

    def _run(self):
        now = time.time()
        next = now + self.period
        while True:
            while next < now:
                next += self.period
            gevent.sleep(next - now)
            self.callback()
            now = time.time()

def get_path(d, path):
    """Get a nested dictionary values, supporting wildcard '*' matches at any level

    >>> sorted(get_path({'a': 1, 'b': 2}, 'a'))
    [('a', 1)]
    >>> sorted(get_path({'a': 1, 'b': 2}, '*'))
    [('a', 1), ('b', 2)]
    >>> sorted(get_path({'a':{'b':{'c': 1}}}, 'a.b.c'))
    [('a.b.c', 1)]
    >>> sorted(get_path({'a': {'b': {'c': 1}, 'd': {'c': 2}}, 'b': {'b': {'c': 3}}}, 'a.*.c'))
    [('a.b.c', 1), ('a.d.c', 2)]
    >>> sorted(get_path({}, 'a.*.c'))
    []
    >>> sorted(get_path({'a': 1}, 'a.b'))
    []
    >>> sorted(get_path({'a': 1}, 'a.*'))
    []
    """

    matches = [([], d)]
    parts = path.split('.')
    for part in parts:
        previous = matches
        matches = []
        
        for key, d in previous:
            if isinstance(d, dict):
                if part == '*':
                    matches.extend((key+[k], v) for k, v in d.iteritems())
                elif isinstance(d, dict) and part in d:
                    matches.append((key+[part], d[part]))

    return [ ('.'.join(k), v) for k, v in matches ]



#!/usr/bin/env python

import sys

__version__ = (1, 0, 4, "final", 0)

def main():
    # avoid importing all dependencies when doing 'import logcabin' for
    # __version__
    from __main__ import LogCabin
    app = LogCabin()
    success = app.run()
    if success:
        sys.exit(0)
    else:
        sys.exit(1)

if __name__ == '__main__':
    main()



from .output import Output
import pymongo

class Mongodb(Output):
    """Outputs to a mongodb collection.

    :param string host: mongodb host
    :param integer port: mongodb port
    :param string database: mongodb database
    :param string collection: mongodb collection
    """

    def __init__(self, host='localhost', port=27017, database='test', collection='events'):
        super(Mongodb, self).__init__()
        self.host = host
        self.port = port
        self.database = database
        self.collection = collection
        self.conn = pymongo.MongoClient(self.host)

    def process(self, event):
        d = dict(event)
        self.conn[self.database][self.collection].insert(d)



from unittest import TestCase
import gevent
from gevent.queue import Queue

from logcabin.event import Event
from logcabin.context import Context, DummyContext
from logcabin.flow import If, Switch, Sequence
from logcabin.filters import mutate

from testhelper import assertEventEquals

class FlowTests(TestCase):
    def create(self, conf={}, events=[]):
        self.input = Queue()
        self.output = Queue()

        context = DummyContext()
        with context:
            self.i = self.create_stage(**conf)
            self.input = self.i.setup(self.output)

        self.assertEquals(1, len(context.stages))

        self.i.start()
        for ev in events:
            self.input.put(ev)
        return self.i

    def wait(self, timeout=1.0, events=1):
        with gevent.Timeout(timeout):
            # wait for input to be consumed and output to be produced
            while self.input.qsize():
                gevent.sleep(0.0)
            while self.output.qsize() < events:
                gevent.sleep(0.0)

        self.i.stop()
        if events:
            return [self.output.get() for n in xrange(events)]

class SequenceTests(FlowTests):
    def create_stage(self):
        with Sequence() as test:
            mutate.Mutate(set={'a': 1})
            mutate.Mutate(set={'b': 2})
        return test

    def test_simple(self):
        self.create({},
            [Event(a=1)])
        q = self.wait()
        assertEventEquals(self, Event(a=1, b=2), q[0])

class SwitchTests(FlowTests):
    def create_stage(self):
        with Switch() as test:
            with test(lambda ev: ev.a == 1):
                mutate.Mutate(set={'b': 1})
            with test('a == 2'):
                mutate.Mutate(set={'b': 2})
            with test.default:
                mutate.Mutate(set={'b': False})
        return test

    def test_lambda(self):
        self.create({},
            [Event(a=1)])
        q = self.wait()
        assertEventEquals(self, Event(a=1, b=1), q[0])

    def test_snippet(self):
        self.create({},
            [Event(a=2)])
        q = self.wait()
        assertEventEquals(self, Event(a=2, b=2), q[0])

    def test_default(self):
        self.create({},
            [Event(a=3)])
        q = self.wait()
        assertEventEquals(self, Event(a=3, b=False), q[0])

    def test_missing(self):
        self.create({},
            [Event()])
        q = self.wait()
        assertEventEquals(self, Event(b=False), q[0])

class IfTests(FlowTests):
    def create_stage(self):
        test = If(lambda ev: ev.a == 1)
        with test:
            mutate.Mutate(set={'b': 1})
        return test

    def test_true(self):
        self.create({},
            [Event(a=1)])
        q = self.wait()
        assertEventEquals(self, Event(a=1, b=1), q[0])

    def test_false(self):
        self.create({},
            [Event(a=2)])
        q = self.wait()
        assertEventEquals(self, Event(a=2), q[0])



from unittest import TestCase
import zmq.green as zmq
import gevent
import gevent.socket as socket
from gevent.queue import Queue
import random
import os

from logcabin.event import Event
from logcabin.context import DummyContext
from logcabin.inputs import udp, zeromq, file as fileinput

from testhelper import TempDirectory, assertEventEquals

class InputTests(TestCase):
    def create(self, conf):
        self.output = Queue()
        with DummyContext():
            self.i = i = self.cls(**conf)
        i.setup(self.output)
        i.start()
        return i

    def waitForQueue(self, timeout=1.0, events=1):
        with gevent.Timeout(timeout):
            while self.output.qsize() < events:
                gevent.sleep(0.0)

        self.i.stop()
        if events:
            return [self.output.get() for n in xrange(events)]

class ZeromqTests(InputTests):
    cls = zeromq.Zeromq

    def test_event(self):
        conf = {'address': 'ipc://testipc'}
        self.create(conf)

        # create a zeromq socket
        ctx = zmq.Context()
        sock = ctx.socket(zmq.PUSH)
        sock.connect(conf['address'])
        sock.send('abc')

        q = self.waitForQueue()
        assertEventEquals(self, Event(data='abc'), q[0])

class UdpTests(InputTests):
    cls = udp.Udp

    def test_event(self):
        conf = {'port': random.randint(1024, 65535)}
        self.create(conf)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.sendto('abc', ('', conf['port']))

        # yield for processing to happen
        q = self.waitForQueue()
        assertEventEquals(self, Event(data='abc'), q[0])

class FileTests(InputTests):
    cls = fileinput.File

    def test_success(self):
        with TempDirectory():
            conf = {'path': 'test*.log'}
            self.create(conf)

            # create log after a short time
            gevent.sleep(0.01)
            with file('test1.log', 'w') as fin:
                print >>fin, 'abc'
                gevent.sleep(0.01)
                print >>fin, 'def'

            q = self.waitForQueue(events=2)
            assertEventEquals(self, Event(data='abc'), q[0])
            assertEventEquals(self, Event(data='def'), q[1])

    def test_rolling(self):
        with TempDirectory():
            with file('test1.log', 'w') as fin:
                print >>fin, 'abc'

            conf = {'path': 'test*.log'}
            self.create(conf)

            gevent.sleep(0.01)
            os.rename('test1.log', 'test1.log.1')
            with file('test1.log', 'w') as fin:
                print >>fin, 'def'

            q = self.waitForQueue(events=2)
            assertEventEquals(self, Event(data='abc'), q[0])
            assertEventEquals(self, Event(data='def'), q[1])

    def test_multiple(self):
        with TempDirectory():
            conf = {'path': 'test*.log'}
            self.create(conf)

            with file('test1.log', 'w') as fin:
                print >>fin, 'abc'
            with file('test2.log', 'w') as fin:
                print >>fin, 'abc'

            q = self.waitForQueue(events=2)
            assertEventEquals(self, Event(data='abc'), q[0])
            assertEventEquals(self, Event(data='abc'), q[1])

    def test_resume(self):
        with TempDirectory():
            with file('test1.log', 'w') as fin:
                print >>fin, 'abc'
            conf = {'path': 'test*.log'}
            self.create(conf)

            q = self.waitForQueue(events=1)
            assertEventEquals(self, Event(data='abc'), q[0])

            with file('test1.log', 'a') as fin:
                print >>fin, 'def'

            self.create(conf)
            q = self.waitForQueue(events=1)
            assertEventEquals(self, Event(data='def'), q[0])

    def test_resume_statedir(self):
        with TempDirectory():
            os.mkdir('state')
            with file('test1.log', 'w') as fin:
                print >>fin, 'abc'
            conf = {'path': 'test*.log', 'statedir': 'state'}
            self.create(conf)

            q = self.waitForQueue(events=1)
            assertEventEquals(self, Event(data='abc'), q[0])

            with file('test1.log', 'a') as fin:
                print >>fin, 'def'

            self.create(conf)
            q = self.waitForQueue(events=1)
            assertEventEquals(self, Event(data='def'), q[0])

    def test_truncated(self):
        with TempDirectory():
            conf = {'path': 'test*.log'}
            self.create(conf)

            with file('test1.log', 'w') as fin:
                print >>fin, 'abc'
                print >>fin, 'def'
            gevent.sleep(0.01)
            with file('test1.log', 'w') as fin:
                print >>fin, 'ghi'

            q = self.waitForQueue(events=3)
            assertEventEquals(self, Event(data='abc'), q[0])

    def test_missing(self):
        with TempDirectory():
            conf = {'path': 'test*.log'}
            self.create(conf)

            self.waitForQueue(events=0)



# avoid naming ambiguity over stdlib json
from __future__ import absolute_import

from .filter import Filter
import json

class Json(Filter):
    """Parse a json encoded field.

    :param string field: the field containing the json (default: data)
    :param boolean consume: whether to remove the field after decoding (default: true)
    """

    def __init__(self, field='data', consume=True, on_error='reject'):
        super(Json, self).__init__(on_error=on_error)
        self.field = field
        self.consume = consume

    def process(self, event):
        if self.field in event:
            try:
                data = event[self.field]
                j = json.loads(data)
                self.logger.debug('JSON decoded: %s' % j)
                if self.consume:
                    del event[self.field]

                event.update(j)
                return True
            except ValueError as ex:
                self._error(event, ex)
                return False
